#ifndef __DATAHANDLE_H__
#define __DATAHANDLE_H__

#include "../std.h"

#include <globus_common.h>
#include <globus_ftp_client.h>
#include <globus_ftp_client_plugin.h>
#include <globus_gass_copy.h>

#include "../misc/condition.h"
#include "../misc/globus_modules.h"

#define MAX_PARALLEL_STREAMS 20

class DataPoint;
class DataBufferPar;
class DataCallback;
class DataPoint::FileInfo;

#ifndef SKIP_HTTPG
class httpg_info_t;
#endif

/*
  Generalization of file handle. 
  Handles upload/download.
*/
class DataHandle {
 public:
  typedef enum {
    common_failure = 0,
    credentials_expired_failure = 1
  } failure_reason_t;
 private:
  typedef enum {
    url_is_unknown,
    url_is_file,
    url_is_ftp,
    url_is_http,
#ifndef SKIP_HTTPG
    url_is_httpg,
#endif
    url_is_bbftp
  } url_proto_t;
  /* set by init_handle */
  url_proto_t url_proto;
  DataPoint* url; /* full url with locations, etc. */
  DataBufferPar* buffer; /* pointer to buffer for data */
  string c_url;   /* current url with options stripped */ 
  bool cacheable;
  bool is_secure;
  bool force_secure;
  bool force_passive;
  bool reading;
  bool writing;
  bool no_checks;
  /* true if allowed to produce scattered data (non-sequetial offsets) */
  bool allow_out_of_order;

  /* ftp */
  GlobusModuleFTPClient ftp_mod;
  bool ftp_active;
  globus_ftp_client_handle_t ftp_handle;
  globus_ftp_client_operationattr_t ftp_opattr;
  globus_ftp_client_plugin_t ftp_plugin;
  globus_thread_t ftp_control_thread;
  int ftp_threads;
  Condition<int> ftp_completed;
  LockSimple ftp_order;
  bool ftp_eof_flag;
  CounterSimple ftp_counter;
  static void ftp_complete_callback(void* arg,globus_ftp_client_handle_t* handle,globus_object_t* error);
  static void ftp_get_complete_callback(void* arg,globus_ftp_client_handle_t* handle,globus_object_t* error);
  static void ftp_put_complete_callback(void* arg,globus_ftp_client_handle_t* handle,globus_object_t* error);
  static void ftp_read_callback(void* arg,globus_ftp_client_handle_t* handle,globus_object_t* error,globus_byte_t* buffer,globus_size_t length,globus_off_t offset,globus_bool_t eof);
  static void ftp_check_callback(void* arg,globus_ftp_client_handle_t* handle,globus_object_t* error,globus_byte_t* buffer,globus_size_t length,globus_off_t offset,globus_bool_t eof);
  static void ftp_write_callback(void* arg,globus_ftp_client_handle_t* handle,globus_object_t* error,globus_byte_t* buffer,globus_size_t length,globus_off_t offset,globus_bool_t eof);
  static void* DataHandle::ftp_read_thread(void *arg);
  static void* DataHandle::ftp_write_thread(void *arg);
  bool start_reading_ftp(DataBufferPar &buf);
  bool start_writing_ftp(DataBufferPar &buf);
  bool stop_reading_ftp(void);
  bool stop_writing_ftp(void);
  bool check_ftp(void);
  bool remove_ftp(void);
  string ftp_dir_path;
  bool mkdir_ftp();
  char ftp_buf[16];
  bool list_files_ftp(list<DataPoint::FileInfo> &files,bool resolve = true);

  /* http */
  GlobusModuleGASSTransfer http_mod;
  bool http_active;
  globus_gass_transfer_request_t http_handle;
  bool http_handle_was_initialized;
  unsigned long long int http_offset;
  bool start_reading_http(DataBufferPar &buffer);
  bool start_writing_http(DataBufferPar &buffer);
  bool stop_reading_http(void);
  bool stop_writing_http(void);
  bool check_http(void);
  static void read_http(void* arg,globus_gass_transfer_request_t handle,globus_byte_t* buffer, globus_size_t length, globus_bool_t last_data);
  static void write_http(void* arg,globus_gass_transfer_request_t handle,globus_byte_t* buffer, globus_size_t length, globus_bool_t last_data);
  static void http_callback(void* arg,globus_gass_transfer_request_t handle);
  static void http_check_callback(void* arg,globus_gass_transfer_request_t handle);

#ifndef SKIP_HTTPG
  /* httpg */
  httpg_info_t* httpg_stat;
  bool start_reading_httpg(DataBufferPar &buffer);
  bool stop_reading_httpg(void);
  bool start_writing_httpg(DataBufferPar &buffer);
  bool stop_writing_httpg(void);
  bool remove_httpg(void);
  bool check_httpg(void);
  bool list_files_httpg(list<DataPoint::FileInfo> &files,bool resolve = true);
#endif

  /* bbftp - no bbftp is not supported */
  bool bbftp_active;

  /* file */
  int fd;
  pthread_t file_thread;
  pthread_attr_t file_thread_attr;
  bool start_reading_file(DataBufferPar &buffer);
  bool start_writing_file(DataBufferPar &buffer,DataCallback *space_cb);
  bool stop_reading_file(void);
  bool stop_writing_file(void);
  bool check_file(void);
  bool remove_file(void);
  static void* read_file(void* arg);
  static void* write_file(void* arg);
  CondSimple file_thread_exited;
  bool list_files_file(list<DataPoint::FileInfo> &files,bool resolve = true);

  failure_reason_t failure_code;

  bool init_handle(void); 
  bool deinit_handle(void); 
 public:
  /* 
    Constructor and destructor
    url_ - generalized url. Should not be destroyed before DataHandle itself 
  */
  DataHandle(DataPoint* url_);
  ~DataHandle(void);
  /* 
     Start reading data from url.
     Separate thread to transfer data will be created.
     buffer - will use this buffer to put information into. Should
       not be destroyed before stop_reading was called.
     returns: true on success. 
  */
  bool start_reading(DataBufferPar &buffer);
  /* 
     Start writing data to url.
     Separate thread to transfer data will be created.
     buffer - will use this buffer to get information from. Should
       not be destroyed before stop_writing was called.
     space_cb - called whicch will be called if there is not enough
       space storing data. Currently implemented only for file:/// url.
     returns: true on success.
  */
  bool start_writing(DataBufferPar &buffer,DataCallback *space_cb = NULL);
  /* stop reading/writing  - these MUST be called after cooressponding
     start_* method, after whole data is transfered. Use 'buffer' to 
     find out when data is transfered.
  */
  bool stop_reading(void);
  bool stop_writing(void);
  /*
    Analyze url and provide some hints.
      bufsize - suggested size of buffers
      bufnum - suggested number of buffers
      cache - true if url can be cached
      local - true if url access lically (file:/// url)
  */
  bool analyze(long int *bufsize,int *bufnum,bool *cache,bool *local);
  /*
    Check if url is accessible.
  */
  bool check(void);
  /*
    Remove url.
  */
  bool remove(void);
  /* 
    list files in directory (url must point to directory/group)
  */
  bool list_files(list<DataPoint::FileInfo> &files,bool resolve = true);

  /* true if can _accept_ scatterd data */
  bool out_of_order(void);
  /* true if allowed to produce scattered data */
  void out_of_order(bool);

  /* true if existance of file/directory should be checked */
  void additional_checks(bool v) { no_checks=!v; };
  bool additional_checks(void) { return !no_checks; };

  bool secure(void);
  void secure(bool);
  void passive(bool);
  failure_reason_t failure_reason(void) { return failure_code; };
};

#endif
