#include "../std.h"
#include <string>

/*
  Keeps track of average and instantaneous speed. Also detects data
  transfer inactivity.
*/

#define DATASPEED_AVERAGING_PERIOD 60

class DataSpeed {
 private:
  time_t first_time;
  time_t last_time;
  time_t last_activity_time;
  unsigned long long int N;
  unsigned long long int Nall;
  time_t first_speed_failure;
  time_t last_printed;

  time_t T;
  time_t min_speed_time;
  time_t max_inactivity_time;
  unsigned long long int min_speed;
  unsigned long long int min_average_speed;
  bool be_verbose;
  string verbose_prefix;

  bool min_speed_failed;
  bool min_average_speed_failed;
  bool max_inactivity_time_failed;

  bool disabled;
 public:
  /*
    Constructors and destructor
    Accepts:
     base - time period used to average values (default 1 minute).
     min_speed_ - minimal allowed speed (B/s). If speed drops and holds
       below for min_speed_time_ seconds error is triggered.
     min_speed_time_
     min_average_speed_ - (B/s) minimal average speed to trigger error.
     max_inactivity_time_ - (s) if no data is passing for specified
       amount of time, error is trigggered.
  */
  DataSpeed(time_t base = DATASPEED_AVERAGING_PERIOD);
  DataSpeed(unsigned long long int min_speed_,time_t min_speed_time_,
            unsigned long long int min_average_speed_,
            time_t max_inactivity_time_,
            time_t base_ = DATASPEED_AVERAGING_PERIOD);
  ~DataSpeed(void);
  /*
    Set/reset printing information about current time speeds, amount of data.
  */
  void verbose(bool val);
  /*
    Print information and add 'prefix' at the beginning of string. 
  */
  void verbose(const string &prefix);
  /*
    Returns: true if speed information is going to be printed.
  */
  bool verbose(void);
  /* 
    Change values of speed control options (optionally) specified during 
    creation of object.
  */
  void set_min_speed(unsigned long long int min_speed_,time_t min_speed_time_);
  void set_min_average_speed(unsigned long long int min_average_speed_);
  void set_max_inactivity_time(time_t max_inactivity_time_);
  void set_base(time_t base_ = DATASPEED_AVERAGING_PERIOD);
  /*
    reset all counters and triggers.
  */
  void reset(void);
  /*
    Inform object, that give amount 'n' of bytes has been transfered.
    To make max_inactivity_time work, application should call this
    method time-to-time.
  */
  bool transfer(unsigned long long int n = 0);
  /*
    If disable=true speed control will be turned off 
  */
  void hold(bool disable);
  /* 
    Functionality:
      Check for triggered errors
    Returns:
      true if correspondng error occured.
  */
  bool min_speed_failure() { return min_speed_failed; };
  bool min_average_speed_failure() { return min_average_speed_failed; };
  bool max_inactivity_time_failure() { return max_inactivity_time_failed; };

  unsigned long long int transfered_size(void) { return Nall; };
};

