/*********************************************************************
 *
 * Authors: Vincenzo Ciaschini - Vincenzo.Ciaschini@cnaf.infn.it 
 *
 * Copyright (c) 2002, 2003 INFN-CNAF on behalf of the EU DataGrid.
 * For license conditions see LICENSE file or
 * http://www.edg.org/license.html
 *
 * Parts of this code may be based upon or even include verbatim pieces,
 * originally written by other people, in which case the original header
 * follows.
 *
 *********************************************************************/

/* $Revision: 1.1 $ */
/* $Date: 2003/05/13 06:58:01 $ */

#ifndef _VOMS_API_H
#define _VOMS_API_H

#include <string>
#include <vector>
#include <openssl/x509.h>

/*! \brief User's characteristics: can be repeated. 
*/
struct data {
  string group; /*!< user's group */
  string role;  /*!< user's role */
  string cap;   /*!< user's capability */
};

/*! \brief The type of data returned.
 */
enum data_type { 
  TYPE_NODATA,  /*!< no data */
  TYPE_STD,     /*!< group, role, capability triplet */
  TYPE_CUSTOM   /*!< result of an S command */
};

struct voms {
  int siglen;         /*!< The length of the VOMS server signature */
  string signature;   /*!< The VOMS server signature */
  string user;        /*!< The user's DN, as from his certificate */
  string userca;      /*!< The CA which signed the user's certificate */
  string server;      /*!< The VOMS server DN, as from its certificate */
  string serverca;    /*!< The CA which signed the VOMS certificate */
  string voname;      /*!< The name of the VO to which the VOMS belongs */
  string uri;         /*!< The URI of the VOMS server */
  string date1;       /*!< Beginning of validity of the user info */
  string date2;       /*!< End of validity of the user info */
  data_type type;     /*!< The type of data returned */
  vector<data> std;   /*!< User's characteristics */
  string custom;      /*!< The data returned by an S command */
};

enum recurse_type { 
  RECURSE_CHAIN, 
  RECURSE_NONE 
};



/*! \brief Error codes.
*/
enum verror_type { 
  VERR_NONE,
  VERR_NOSOCKET,  /*!< Socket problem*/
  VERR_NOIDENT,   /*!< Cannot identify itself (certificate problem) */
  VERR_COMM,      /*!< Server problem */
  VERR_PARAM,     /*!< Wrong parameters*/
  VERR_NOEXT,     /*!< VOMS extension missing */
  VERR_NOINIT,    /*!< Initialization error */
  VERR_TIME,      /*!< Error in time checking */
  VERR_IDCHECK,   /*!< User data in extension different from the real ones */
  VERR_EXTRAINFO, /*!< VO name and URI missing */
  VERR_FORMAT,    /*!< Wrong data format */
  VERR_NODATA,    /*!< Empty extension */
  VERR_PARSE,     /*!< Parse error */
  VERR_DIR,       /*!< Directory error */
  VERR_SIGN,      /*!< Signature error */
  VERR_SERVER,    /*!< Unidentifiable VOMS server */
  VERR_MEM,       /*!< Memory problems */
  VERR_VERIFY,    /*!< Generic verification error*/
  //  VERR_IDENT, 
  VERR_TYPE       /*!< Returned data of unknown type */
};


struct vomsdata {
  private:
  string ca_cert_dir;
  string voms_cert_dir;

  public:
  verror_type error; /*!< Error code */

  vomsdata(string voms_dir = "", 
	   string cert_dir = ""); /*!< \param voms_dir The directory which contains the certificate 
                                              of the VOMS server
                                       \param cert_dir The directory which contains the certificate of the CA

                                       If voms_dir is empty, the value of the environment variable
                                       X509_VOMS_DIR is taken

                                       If cert_dir is empty, the value of the environment variable
                                       X509_CERT_DIR is taken
                                   */

  bool Retrieve(X509 *cert, STACK_OF(X509) *chain, 
		recurse_type how = RECURSE_CHAIN); /*!< Extracts the VOMS extension from an X.509 certificate. 

                                                     The function doesn't check the validity of the certificates, 
                                                     but it does check the content of the user data.
                                                     \param cert The certificate with the VOMS extensions
                                                     \param chain The chain of the validation certificates 
                                                           (only the intermediate ones)
                                                     \param how Recursion type
                                                     \return failure (F) or success (T)
                                                   */  
  bool Contact(string hostname, 
               int port, 
               string servsubject, 
               string command); /*!< Contacts a VOMS server to get a certificate

                                    It is the equivalent of the voms_proxy_init command, but 
                                    without the --include functionality.
                                    \param hostname FQDN of the VOMS server
                                    \param port the port on which the VOMS server is listening
                                    \param servsubject the subject of the server's certificate
                                    \param command Command
                                    \return failure (F) or success (T)
				*/
  bool Import(string buffer);/*!< Converts data from the format used for inclusion 
                                  into a certificate to the internal format

                                     The function does verify the data.
                                     \param buffer contains the data to be converted 
                                     \return Failure (F) or Success (T)
                             */
  bool Export(string &data); /*!< Exports data from vomsdata::data to the format 
                                  used for inclusion into a certificate. 

                                  The function doesn't verify the data
                                  \param data The certificate extension
                                  \return Failure (F) or Success (T)
                             */

  vector<voms> data; /*!< User's info, as in the certificate extension.
                          It may contain data gathered from more than one VOMS server, 
                     */
  string workvo;     /*!< The value of the -vo option of the voms-proxy-init command */
  string extra_data; /*!< The data specified by the user with the --include switch.
 
                          Note that this field doesn't contain the result of a request
                          to the VOMS server, but instead data specified by the user.

                          The reason for the introduction of this extension is to let 
                          a user include important data into his proxy certificate, 
                          like, for example, a kerberos ticket 
                      */

};


int getMajorVersionNumber(void);
int getMinorVersionNumber(void);
int getPatchVersionNumber(void);

#endif
