/*********************************************************************
 *
 * Authors: Vincenzo Ciaschini - Vincenzo.Ciaschini@cnaf.infn.it 
 *
 * Copyright (c) 2002, 2003 INFN-CNAF on behalf of the EU DataGrid.
 * For license conditions see LICENSE file or
 * http://www.edg.org/license.html
 *
 * Parts of this code may be based upon or even include verbatim pieces,
 * originally written by other people, in which case the original header
 * follows.
 *
 *********************************************************************/
#include <openssl/pem.h>
#include <openssl/asn1.h>
#include <openssl/evp.h>

using namespace std;

#ifdef _XOPEN_SOURCE
#undef _XOPEN_SOURCE
#endif
#ifdef __USE_XOPEN
#undef __USE_XOPEN
#endif

#include <iostream>
#include <iomanip>

#include "data.h"
#include "format.h"
#include "sign.h"

static bool parse(string &is, const string s);
static bool parse(string &is, const string s, int &val);
static bool parse(string &is, const string s, string &val);

/*
 * Function:
 *   get_number(str, name, limit)
 *
 * Description:
 *   This function extracts from the buffer pointed by *'str' a non-zero,
 *   newline-terminated numeric parameter that has the 'name' string has
 *   a prefix.
 *
 * Parameters:
 *   'str'   - A pointer to a pointer to the buffer.  This is done because
 *             the pointer itself will be adjusted in case of successful
 *             execution.
 *   'name'  - The prefix of the number.
 *   'limit' - A pointer to the limit of the buffer.
 *
 * Result:
 *   Failure:
 *     0
 * Success:
 *   The numeric value found. Furthermore, *'str' will be adjusted to point
 *   immediately after the data examined by the function.
 */
static bool
parse(string &is, const string s, int &val)
{
  int len = s.length();
  string v = is.substr(0, len);

  if (v == s) {
    val = 0;

    while(isdigit(is[len])) {
      val *= 10;
      val += is[len++]- '0';
    }
    if (is[len++] == '\n') {
      is.erase(0, len);
      return true;
    }
  }

  return false;
}

static bool  
parse(string &is, const string s, string &val)
{
  int len = s.length();
  string v = is.substr(0, len);

  if (v == s) {
    string::size_type pos = is.find('\n');
    if (pos != string::npos) {
      val = string(is.substr(len, pos-len));
      is.erase(0,pos+1);
      return true;
    }
  }

  return false;
}

static bool  
parse(string &is, const string s)
{
  int len = s.length();
  string v = is.substr(0, len);

  if (v == s) {
    is.erase(0, len);
    return true;
  }

  return false;
}

/*
 * Function:
 *   unformat(data, len, key)
 *
 * Description:
 *   This function takes a buffer full of data returned by the format()
 *   (See) function, analyzes it for general correctness and rescinds it
 *   into its components, that are subsequently memorized into a collection
 *   structure.
 *
 * Parameters:
 *   'data' - Pointer to the buffer to analyze.
 *   'len'  - Length of the buffer.
 *   'key'  - Pointer to the server's public key.
 *
 * Result:
 *   Failure:
 *     NULL.
 *   Success:
 *     A pointer to a struct collection containing the broken down data
 *     that must be manually freed.  To do so, there is no need to also
 *     free its pointed data.
 */
bool
unformat(const string data, const EVP_PKEY *key, struct collection &results)
{
  if (data.empty() || !key)
    return false;

  /*
   * Load the signature.
   */

  int sig_len = 0;
  bool ok;

  string is=data;

  ok  = parse(is, "SIGLEN:", sig_len);
  ok &= parse(is, "SIGNATURE:");

  if (!ok)
    return false;

  string signature = is.substr(0,sig_len);
  string signed_data = is.substr(sig_len);

#ifdef DEBUG
  cout << "SIGNED_DATA: " << signed_data << endl << flush;
#endif

  /*
   * Verifys it.
   */
  if (!verify(const_cast<EVP_PKEY *>(key), signed_data, signature))
    return false;

  is = signed_data;

  /*
   * Extracts data from the buffer.
   */

  ok  = parse(is, "USER:", results.user);
  ok &= parse(is, "UCA:", results.userca);
  ok &= parse(is, "SERVER:", results.server);
  ok &= parse(is, "SCA:", results.serverca);
  ok &= parse(is, "VO:", results.voname);
  ok &= parse(is, "URI:", results.uri);
  ok &= parse(is, "TIME1:" , results.date1);
  ok &= parse(is, "TIME2:", results.date2);
  ok &= parse(is, "DATALEN:", results.datalen);

  if (results.user.empty() || results.server.empty() ||
      results.date1.empty() || results.date2.empty() ||
      results.datalen == 0)
    return false;

  results.data = is;

  return ok;
}

/*
 * Function:
 *   format(data, user, userca, server, serverca, key, valid)
 *
 * Description:
 *   This function formats a n-tuple of data to be transmitted from the
 *   server to the client.
 *
 * Parameters:
 *   'data'     - A string containing the data to be transmitted.
 *   'user'     - The subject of the users's certificate.
 *   'userca'   - The subject of the CA that issued the user's certificate.
 *   'server'   - The subject of the server's certificate.
 *   'serverca' - The subject of the CA that issued the server's certificate.
 *   'key'      - The private key to use for signing.
 *   'valid'    - How long this information should be valid.
 *   'voname'   - The name of the VO that owns this server.
 *   'result'   - The formatted data.
 *
 * Result:
 *   A boolean indicating success or failure.
 */
bool
format(const string data, const string user, const string userca, 
       const string server, const string serverca, const EVP_PKEY *key,
       int valid, const string voname, const string uri, string &result)
{
  string temp;
  string signature;
  unsigned char buf1[30];
  unsigned char *bufp1 = buf1;
  unsigned char buf2[30];
  unsigned char *bufp2 = buf2;

  if (data.empty() || user.empty() || server.empty() ||
      !key || userca.empty() || voname.empty() ||
      serverca.empty())
    return false;


  //cerr << "FORMAT: " << endl << "  data: " << data << endl << flush;
  /*
   * Creates a representation of ASN1 validity dates
   */
  ASN1_TIME *time1 = ASN1_TIME_new();
  ASN1_TIME *time2 = ASN1_TIME_new();
  if (!time1 || !time2)
    goto error;

  time_t curtime;
  time(&curtime);
  time1 = ASN1_TIME_set(time1, curtime);
  time2 = ASN1_TIME_set(time2, curtime);
  time2 = X509_time_adj(time2, valid, &curtime);


  if ((i2d_ASN1_TIME(time1, &bufp1) == -1)  || (i2d_ASN1_TIME(time2, &bufp2) == -1))
    goto error;

  *bufp1='\0';
  *bufp2='\0';

  /*
   * Adds data to a return blob.
   */
  result = string("USER:") + user + "\nUCA:" + userca + "\nSERVER:" + server +
    "\nSCA:" + serverca + "\nVO:" + voname + "\nURI:" + uri + "\nTIME1:" +
    (char *)buf1 + "\nTIME2:" + (char *)buf2 + "\nDATALEN:" +
    stringify(data.size(),temp) + "\n" + data;

  /*
   * Creates a signature of the data
   */
  if (!sign(const_cast<EVP_PKEY *>(key), result.c_str(), signature))
    goto error;

  /*
   * Adds the signature
   */
  result = string("SIGLEN:") + stringify(signature.size(),temp) +
    + "\nSIGNATURE:" + signature + result;

  ASN1_TIME_free(time1);
  ASN1_TIME_free(time2);

  return true;

error:
  if (time1)
    ASN1_TIME_free(time1);
  if (time2)
    ASN1_TIME_free(time2);

  return false;
}
