#include "../../std.h"

#include <iostream>
#include <fstream>

#include "../../misc/stringtoint.h"
#include "../../misc/escaped.h"
#include "../../misc/time_utils.h"
#include "file.h"


// ----------------------------------------------------------

const char* file_state_str[FILE_STATE_MAX] = {
  "accepted",
  "collecting",
  "complete",
  "deleting",
};
const char* reg_state_str[REG_STATE_MAX] = {
  "local",
  "registering",
  "announced",
  "unregistering",
};

bool set_state(const char* key,const char* value,void* arg) {
  SEState* s = (SEState*)arg;
  return s->set(key,value);
}

bool SEState::set(const char* key,const char* value) {
  if(strcasecmp(key,"file") == 0) {
    if(value == NULL) return false;
    int i;
    for(i=0;i<FILE_STATE_MAX;i++) if(strcasecmp(value,file_state_str[i]) == 0) break;
    if(i>=FILE_STATE_MAX) return false;
    file_=i;
  } else if(strcasecmp(key,"registration") == 0) {
    if(value == NULL) return false;
    int i;
    for(i=0;i<FILE_STATE_MAX;i++) if(strcasecmp(value,reg_state_str[i]) == 0) break;
    if(i>=REG_STATE_MAX) return false;
    reg_=i;
  };
  return true;
}

ostream& operator<<(ostream& o,const SEState &s) {
  o<<"file="<<file_state_str[s.file_]<<endl;
  o<<"registration="<<reg_state_str[s.reg_]<<endl;
  return o;
}

// ----------------------------------------------------------

typedef bool(*use_pair_t)(const char* key,const char* value,void* arg);

bool read_pairs(const char* fname,use_pair_t func,void* arg) {
  ifstream f(fname);
  if(!f.is_open() ) return false;
  char buf[1024];
  for(;!f.eof();) {
    istream_readline(f,buf,sizeof(buf));
    char* key = buf;
    for(;(*key) && isspace(*key);key++);
    if((*key) == '#') continue;
    char* value = strchr(key,'='); 
    if(value) { (*value)=0; value++; };
    if(!func(buf,value,arg)) return false;
  };
  return true;
}

// ----------------------------------------------------------

static int write_range(const char* fname,SEFileRange* ranges) {
  if(ranges == NULL) {
    unlink(fname); return 0;
  };
  ofstream o(fname,ios::trunc);
  if(!o) return -1;
  for(int i=0;i<MAX_SEFILE_RANGES;i++) {
    if(ranges[i].start == (uint64_t)(-1)) continue;
    o<<ranges[i].start<<" "<<ranges[i].end<<endl;
    if(!o) {

    };
  };
  return 0;
}

static int read_range(const char* fname,SEFileRange* ranges) {
  struct stat st;
  if(stat(fname,&st) != 0) {
    if(errno == ENOENT) {
      ranges=NULL;
      return 1;
    };
  };
  int i;
  for(i=0;i<MAX_SEFILE_RANGES;i++) ranges[i].start=(uint64_t)(-1);
  uint64_t start; 
  uint64_t end;
  ifstream in(fname);
  if(!in) return -1;
  i=0;
  for(;i<MAX_SEFILE_RANGES;i++) {
    if(in.eof()) break;
    in>>start; if(!in) return -1;
    in>>end; if(!in) return -1;
    in.ignore(256,'\n');
    ranges[i].start=start; ranges[i].end=end;
  };
  return 0;
}

static int join_range(uint64_t start, uint64_t end,SEFileRange* ranges) {
  int i;
  int empty_slot = MAX_SEFILE_RANGES;
  for(i=0;i<MAX_SEFILE_RANGES;i++) {
    if(ranges[i].start == (uint64_t)(-1)) { empty_slot=i; continue; };
    if((ranges[i].start >= start) && (ranges[i].start <= end)) break;
    if((ranges[i].end >= start) && (ranges[i].end <= end)) break;
  };
  if(i<MAX_SEFILE_RANGES) {
    uint64_t min=ranges[i].start;
    uint64_t max=ranges[i].end;
    if(start < ranges[i].start) ranges[i].start=start;
    if(end > ranges[i].end) ranges[i].end=end;
  } else {
    if(empty_slot == MAX_SEFILE_RANGES) return -1;
    i=empty_slot;
    ranges[i].start=start;
    ranges[i].end=end;
  };
  return i;
}

static int compress_range(SEFileRange* ranges) {
  int i;
  int n = 0;
  for(i=0;i<(MAX_SEFILE_RANGES-1);i++) {
    if(ranges[i].start == (uint64_t)(-1)) continue;
    int r = join_range(ranges[i].start,ranges[i].end,ranges+i+1);
    if(r != -1) { // range replaced
      n++;
      ranges[i].start=(uint64_t)(-1);
      ranges[i].end=0;
    };
  };
  if(ranges[i].start != (uint64_t)(-1)) n++;
  return n;
}

// ----------------------------------------------------------

int SEFile::open(bool for_read) {
  int i;
  lock.block();
  if(for_read) {
    if(ranges != NULL) { lock.unblock(); return -1; }; // incomplete
    if(file_handle==-1) file_handle=::open64(path.c_str(),O_RDWR);
    if(file_handle==-1) { lock.unblock(); return -1; }; // failed to open
    read_count++;
    lock.unblock();
    return 0;
  };
  if((write_count >= (MAX_SEFILE_RANGES-2)) && (ranges != NULL)) { 
    lock.unblock(); return -1; // too scattered
  };
  if(file_handle==-1) file_handle=::open64(path.c_str(),O_RDWR);
  if(file_handle==-1) { lock.unblock(); return -1; }; // failed to open
  write_count++;
  lock.unblock();
  return 0;
}

bool SEFile::state(int f,int r) {
  lock.block();
  if(((f<0) || (f>=FILE_STATE_MAX)) && ((r<0) || (r>=FILE_STATE_MAX))) {
    lock.unblock(); return false;
  };
  string sname = path+".state";
  ofstream o(sname.c_str(),ios::trunc);
  if(!o) { lock.unblock(); return false; };
  state_.file(f); state_.reg(r); o<<state_;
  if(!o) { lock.unblock(); return false; }
  lock.unblock();
  return true;
}

void SEFile::close(bool for_read) {
  lock.block();
  if(for_read) { read_count--; }
  else { write_count--; };
  if((read_count==0) && (write_count==0)) {
    ::close(file_handle); file_handle=-1;
  };
  // join ranges
  if(ranges && (!for_read)) {
    int n = compress_range(ranges);
    if(n == 1) {
      // check if full file
      if((ranges[MAX_SEFILE_RANGES-1].start == 0) &&
         (ranges[MAX_SEFILE_RANGES-1].end >= size())) {
        cerr<<"SEFile::close: file is full (length = "<<ranges[MAX_SEFILE_RANGES-1].end<<")"<<endl;
        free(ranges); ranges=NULL;
      };
    };
    // store updated range information
    string rname = path+".range";
    if(write_range(rname.c_str(),ranges) == -1) {

    };
  };
/*
  // update state
  if(complete()) {
    string sname = path+".state";
    ofstream o(sname.c_str(),ios::trunc);
    state_.file(FILE_STATE_COMPLETE); o<<state_;
    if(!o) {    }
  } else {
    cerr<<"SEFile::close: file attributes are not complete yet"<<endl;
  };
*/
  lock.unblock();
  return;
}

void SEFile::destroy(void) {
  valid=false;
cerr<<"SEFile::destroy: unlink "<<path+".range"<<endl;
  unlink((path+".range").c_str());
cerr<<"SEFile::destroy: unlink "<<path+".attr"<<endl;
  unlink((path+".attr").c_str());
cerr<<"SEFile::destroy: unlink "<<path+".state"<<endl;
  unlink((path+".state").c_str());
cerr<<"SEFile::destroy: unlink "<<path+".acl"<<endl;
  unlink((path+".acl").c_str());
cerr<<"SEFile::destroy: unlink "<<path<<endl;
  unlink(path.c_str());
//  state_=FILE_STATE_NONEXISTENT;
}

static char char_tab[16] = {
   '0','1','2','3',
   '4','5','6','7',
   '8','9','a','b',
   'c','d','e','f'
};

static void file_name(uint32_t n,char* name) {
  for(int i=(sizeof(uint32_t)*2-1);i>=0;i--) {
    name[i] = char_tab[n & 0x0F]; n>>=4;
  };
  name[sizeof(int)*2]=0;
}

/* creating new file */
SEFile::SEFile(const char* dirpath,const SEAttributes& attr):path(dirpath),SEAttributes(attr) {
//  id=id_; // TODO check for same name ?
cerr<<"SEFile::SEFile(new): path: "<<path<<endl;
  char fname[sizeof(uint32_t)*2+2]; fname[0]='/';
  unsigned int i=0;
  srandom(time(NULL));
  unsigned int r = random();
  for(;i<INT_MAX;i++) {
    file_name(i+r,fname+1);
    /* Warning: O_EXCL is broken on NFS */
    string fpath=path+fname; 
    int nh=::open(fpath.c_str(),O_CREAT | O_EXCL | O_RDWR,S_IRUSR | S_IWUSR);
    if(nh == -1) {
      if(errno == ENOSPC) { /* out of space */
        i=INT_MAX; break;
      };
      continue;
    };
    close(nh);
    break;
  };
  if(i == INT_MAX) {


  };
  path+=fname; 
  name=path.c_str()+strlen(dirpath)+1;
  // create file
  int h = ::open64(path.c_str(),O_WRONLY | O_CREAT);
  if(h == -1) {


  };
  close(h);
  // initialize ranges
  ranges=(SEFileRange*)malloc(sizeof(SEFileRange)*MAX_SEFILE_RANGES);
  for(int i = 0;i<MAX_SEFILE_RANGES;i++) ranges[i].start=(uint64_t)(-1);
  string rname = path+".range";
  if(write_range(rname.c_str(),ranges) == -1) {

  };
  rname = path+".attr";
  if(SEAttributes::write(rname.c_str()) != 0) {


  };
  if(!state(FILE_STATE_ACCEPTED,REG_STATE_UNCHANGED)) {


  };

  file_handle=-1; read_count=0; write_count=0;
}


/* acquire existing file */
SEFile::SEFile(const char* path_):path(path_) {
cerr<<"SEFile::SEFile: path: "<<path<<endl;
  valid=false;
  string tpath = path+".attr";
  if(SEAttributes::read(tpath.c_str()) != 0) {


  };
  int n = path.rfind('/');
  if(n == string::npos) { n=0; } else { n++; };
  name=path.c_str()+n;
cerr<<"file: "<<name<<" - "<<id()<<endl;
  // initialize ranges
  tpath = path+".range";
  ranges=(SEFileRange*)malloc(sizeof(SEFileRange)*MAX_SEFILE_RANGES);
  int err;
  if((err=read_range(tpath.c_str(),ranges)) != 0) {
    if(err == 1) { free(ranges); ranges=NULL; }
    else {

    };
  };
  tpath = path+".state";
  if(!read_pairs(tpath.c_str(),&set_state,&state_)) {


  };
  if(state_.reg() == REG_STATE_REGISTERING) {
    state(FILE_STATE_UNCHANGED,REG_STATE_LOCAL);
    cerr<<"Warning: intermidiate registration state REGISTERING found. Setting to LOCAL."<<endl;
  } else if(state_.reg() == REG_STATE_UNREGISTERING) {
    state(FILE_STATE_UNCHANGED,REG_STATE_ANNOUNCED);
    cerr<<"Warning: intermidiate registration state UNREGISTERING found. Setting to ANNOUNCED."<<endl;
  };
  file_handle=-1; read_count=0; write_count=0;
  valid=true;
}




SEFile::~SEFile(void) {
cerr<<"SEFile::~SEFile"<<endl;
  lock.block();
  if(ranges) compress_range(ranges);
  // update ranges information 

  if(ranges) free(ranges); ranges=NULL;
  if(file_handle != -1) close(file_handle); file_handle=-1;
  lock.unblock();
}


uint64_t SEFile::write(void* buf,uint64_t offset,uint64_t size) {
cerr<<"SEFile::write"<<endl;
  if(size==0) return 0;
  if(ranges == NULL) return size; // full file is already available
  // store data
  uint64_t o = offset;
  uint64_t s = size;
  for(;;) {
    ssize_t l = pwrite(file_handle,buf,s,o);
    if(l == -1) return 0;
    s-=l;
    if(s == 0) break;
    o+=l; ((char*)buf)+=l;
  };
  lock.block();
  if(ranges) {
    if(join_range(offset,offset+size,ranges) == -1) { // file is too fragmented
      lock.unblock();
      return 0;
    };
  };
  lock.unblock();
  return size;
}

uint64_t SEFile::read(void* buf,uint64_t offset,uint64_t size) {
cerr<<"SEFile::read - size: "<<size<<endl;
cerr<<"SEFile::read - offset: "<<offset<<endl;
  uint64_t o = offset;
  uint64_t s = size;
  for(;;) {
cerr<<"SEFile::read - pread: "<<file_handle<<","<<s<<","<<o<<endl;
    ssize_t l = pread(file_handle,buf,s,o);
cerr<<"SEFile::read - l: "<<l<<endl;
    if(l == -1) {
      perror("pread");
      return 0;
    };
    if(l == 0) break;
    s-=l;
    if(s == 0) break;
    o+=l; ((char*)buf)+=l;
  };
  return (size-s);
}


SEFileHandle::SEFileHandle(SEFile& f,uint64_t o,bool for_read):file(f) {
  active=true;
  offset=o;
  open_read=for_read;
  if(file.open(open_read) != 0) active=false;
cerr<<"SEFileHandle constructor: active: "<<active<<endl;
}

SEFileHandle::~SEFileHandle(void) {
  if(active) file.close(open_read);
}

uint64_t SEFileHandle::read(void* buf,uint64_t size) {
cerr<<"SEFileHandle::read - size: "<<size<<endl;
  if(!active) return 0;
  uint64_t l = file.read(buf,offset,size);
  offset+=l;
cerr<<"SEFileHandle::read - l: "<<l<<endl;
  return l;
}

uint64_t SEFileHandle::write(void* buf,uint64_t size) {
  if(!active) return 0;
  uint64_t l = file.write(buf,offset,size);
  offset+=l;
  return l;
}

SEFileHandle* SEFileHandle::open(const char* id,uint64_t o,bool for_read,const SEFiles& files) {
  SEFile* file = files.get(id);
  if(!file) return NULL;
  SEFileHandle* f = new SEFileHandle(*file,o,for_read);
  if(!(*f)) { delete f; return NULL; };
  return f;
}

SEFiles::SEFiles(void) {
cerr<<"SEFiles::SEFiles(void)"<<endl;
  files=NULL;
  ns=NULL;
  nfiles=0;
  valid=false;
  reg_type=registration_immediate;
}

SEFiles::SEFiles(const char* dirpath):path(dirpath) {
cerr<<"SEFiles::SEFiles(const char* dirpath)"<<endl;
  valid=false;
  files=NULL;
  ns=NULL;
  reg_type=registration_immediate;
  nfiles=0;
  struct dirent file_;
  struct dirent *file;
  DIR *dir=opendir(dirpath);
  if(dir == NULL) {
    cerr<<"Failed opening directory: "<<dirpath<<endl; return;
  };
  for(;;) {
    readdir_r(dir,&file_,&file);
    if(file == NULL) break;
    int l=strlen(file->d_name);
    if(l>5) {
      if(!strcmp(file->d_name+l-5,".attr")) {
        string name(file->d_name); name.resize(l-5);
        name="/"+name; name=dirpath+name;
cerr<<"SEFiles: creating SEFile: "<<name<<endl;
        SEFile* f = new SEFile(name.c_str());
cerr<<"SEFiles: adding SEFile: "<<name<<endl;
        add(*f);
cerr<<"Added file: "<<f->id()<<endl;
      };
    };
  };
  closedir(dir);
  valid=true;
}

SEFiles::~SEFiles() {
  lock.block();
  for(int i=0;i<nfiles;i++) {
    if(files[i]) delete files[i];
  };
  free(files); files=NULL; valid=false;
  lock.unblock();
}

SEFile* SEFiles::add(SEFile& file) {
cerr<<"SEFiles::add"<<endl;
  lock.block();
  if(files) {
    SEFile* f = find_by_id(file.id());
    if(f) {
      cerr<<"SEFiles::add: file already exists: "<<file.id()<<endl;
      file.destroy();
      lock.unblock();
      return NULL;
    };
    for(int i=0;i<nfiles;i++) {
      if(!files[i]) {
        cerr<<"SEFiles::add: new file: "<<file.id()<<endl;
        files[i]=&file;
        f=files[i];
        lock.unblock();
        return f;
      };
    };
  };
  SEFile** f = (SEFile**)realloc(files,sizeof(SEFile*)*(nfiles+10));
  if(f == NULL) {
    cerr<<"SEFiles::add: failed to allocate memory"<<endl;
    file.destroy();
    lock.unblock();
    return NULL;
  };
  files=f;
  int i=nfiles; files[i]=&file;
  int n=i; i++;
  nfiles+=10; for(;i<nfiles;i++) files[i]=NULL;
  cerr<<"SEFiles::add: new file: "<<file.id()<<endl;
  SEFile* f_ = files[n]; 
  lock.unblock();
  return f_;
}

int SEFiles::remove(SEFile& file) {
cerr<<"SEFiles::remove: "<<file.id()<<endl;
  lock.block();
  for(int i=0;i<nfiles;i++) {
    if(files[i] == &file) {
cerr<<"SEFiles::remove: removing from list"<<endl;
      files[i]=NULL;
    };
  };
cerr<<"SEFiles::remove: calling destroy"<<endl;
  file.destroy(); 
  lock.unblock();
  return 0;
}

SEFile* SEFiles::find_by_id(const char* id) const {
  for(int i=0;i<nfiles;i++) {
    if(files[i]) if(files[i]->check_id(id)) {
      SEFile* f = files[i];
      return f;
    };
  };
  return NULL;
}

SEFile* SEFiles::get(const char* id) const {
  ((LockSimple*)(&lock))->block();
  SEFile* f = find_by_id(id);
  ((LockSimple*)(&lock))->unblock();
  return f;
}

// Register all files, which require registration
int SEFiles::Register(void) {
cerr<<"SEFiles::Register"<<endl;
  lock.block();
  int failures = 0;
  if(files && ns) {
cerr<<"SEFiles::Register: have both files and NS"<<endl;
    for(SEFile** f = files;*f!=NULL;f++) {
      if(*f) {
cerr<<"SEFiles::Register: have valid file: "<<(*f)->id()<<endl;
        ns->Connect(); // take advantage of possibly less reconnections
        // check if file registered
cerr<<"SEFiles::Register: reg state: "<<(*f)->state_reg()<<endl;
cerr<<"SEFiles::Register: retry: "<<register_retry()<<endl;
        if( ((*f)->state_reg()==REG_STATE_LOCAL) && register_retry() ) {
cerr<<"SEFiles::Register: immediate: "<<register_immediately()<<endl;
cerr<<"SEFiles::Register: state: "<<(*f)->state()<<endl;
          if(  register_immediately() ||
               ((*f)->state()==FILE_STATE_COMPLETE)  ) {
            // register new file
cerr<<"Registering: "<<(*f)->id()<<endl;
            if((*f)->state(FILE_STATE_UNCHANGED,REG_STATE_REGISTERING)) {
              if(ns->Register(**f) != 0) {
                failures++;
                (*f)->state(FILE_STATE_UNCHANGED,REG_STATE_LOCAL);
              } else {
                (*f)->state(FILE_STATE_UNCHANGED,REG_STATE_ANNOUNCED);
              };
            };
          };
        };
      };
    };
    ns->Disconnect();
  };
  lock.unblock();
  return failures;
}

// Unregister all files being deleted
int SEFiles::Unregister(void) {
cerr<<"SEFiles::Unregister"<<endl;
  lock.block();
  int failures = 0;
  if(files && ns) {
cerr<<"SEFiles::Unregister: have both files and NS"<<endl;
    for(SEFile** f = files;*f!=NULL;f++) {
      if(*f) {
cerr<<"SEFiles::Unregister: have valid file: "<<(*f)->id()<<endl;
        ns->Connect(); // take advantage of possibly less reconnections
        // check if file registered
cerr<<"SEFiles::Unregister: state: "<<(*f)->state()<<endl;
cerr<<"SEFiles::Unregister: reg state: "<<(*f)->state_reg()<<endl;
cerr<<"SEFiles::Unregister: retry: "<<register_retry()<<endl;
        if( ((*f)->state()==FILE_STATE_DELETING) && register_retry() ) {
          if((*f)->state_reg()!=REG_STATE_UNREGISTERING) {
            // unregister file
cerr<<"Unregistering: "<<(*f)->id()<<endl;
            if((*f)->state(FILE_STATE_UNCHANGED,REG_STATE_UNREGISTERING)) {
cerr<<"Unregistering - calling Unregister"<<endl;
              if(ns->Unregister(**f) != 0) {
cerr<<"Unregistering - Unregister failed"<<endl;
                failures++;
                (*f)->state(FILE_STATE_UNCHANGED,REG_STATE_ANNOUNCED);
              } else {
cerr<<"Unregistering - Unregister succeeded"<<endl;
                // remove(**f);
                (*f)->destroy();
                (*f)=NULL; 
              };
            };
          };
        };
      };
    };
  };
  lock.unblock();
  return failures;
}

int SEFile::write_attr(void) {
  string rname = path+".attr";
  rname = path+".attr";
cerr<<"SEFile::write_attr: to file: "<<rname<<endl;
  return SEAttributes::write(rname.c_str());
}


// -------------------------------------------------------------

#define skip_space(__p) for(;*__p;__p++) if(!isspace(*__p)) break;
#define skip_till_space(__p) for(;*__p;__p++) if(isspace(*__p)) break;

SEAttributes::SEAttributes(const char* fname) {
  valid=false;
  size_b=true;
  checksum_b=false;
  created_b=false;
  read(fname);
}

int SEAttributes::read(const char* fname) {
cerr<<"SEAttributes::read"<<endl;
  ifstream f(fname);
  if(!f) return -1;
  char buf[1024];
  bool have_id = false;
  bool have_size = false;
  for(;!f.eof();) {
    char buf[1024];
    istream_readline(f,buf,sizeof(buf));
    char* p = buf;
    skip_space(p);
    char* command = p;
    if(*command == '#') continue;
cerr<<"SEAttributes::read: line "<<p<<endl;
    skip_till_space(p);
    int command_len = p-command;
    if(!command_len) continue;
    if((command_len==4) && (strncmp(command,"size",command_len)==0)) {
cerr<<"SEAttributes::read: command size: "<<p<<endl;
      string value(p);
      if(!stringtoint(value,size_i)) {
        cerr<<"Size value is bad"<<endl; return -1;
      } else { have_size=true; };
    } else if((command_len==2) && (strncmp(command,"id",command_len)==0)) {
cerr<<"SEAttributes::read: command id: "<<p<<endl;
      if(input_escaped_string(p,id_i) == 0) {
        cerr<<"Can't read ID"<<endl; return -1;
      } else { have_id=true; };
    } else if((command_len==8) && (strncmp(command,"checksum",command_len)==0)){
cerr<<"SEAttributes::read: command checksum: "<<p<<endl;
      if(input_escaped_string(p,checksum_i) == 0) {	 
        cerr<<"Can't read checksum"<<endl; return -1;
      } else { checksum_b=true; };
    } else if((command_len==7) && (strncmp(command,"creator",command_len)==0)){
cerr<<"SEAttributes::read: command creator: "<<p<<endl;
      if(input_escaped_string(p,creator_i) == 0) {
        cerr<<"Can't read icreator"<<endl; return -1;
      };
    } else if((command_len==7) && (strncmp(command,"created",command_len)==0)){
cerr<<"SEAttributes::read: command created: "<<p<<endl;
      string created_str;
      if(input_escaped_string(p,created_str) == 0) {	 
        cerr<<"Can't read creation time"<<endl; return -1;
      } else {
cerr<<"SEAttributes::read: command created: time string: "<<created_str<<endl;
        if(stringtotime(created_i,created_str)) {
          cerr<<"Can't interpret creation time"<<endl; return -1;
        };
        created_b=true;
      };
    };
  };
  if(!have_id)   { cerr<<"Missing ID"<<endl;   return -1; };
  if(!have_size) { cerr<<"Missing size"<<endl; return -1; };
  valid=true;
  return 0;
}

void SEAttributes::created(const time_t* c) {
  struct tm* t;
  struct tm tt;
  created_b=false;
  t=gmtime_r(c,&tt);
  if(!t) return;
  memcpy(&created_i,t,sizeof(struct tm));
  created_b=true;
}

void SEAttributes::created(const char* c) { 
  created_b=false;
  if(!c) return;
  string s(c);
  if(stringtotime(created_i,s)) {
    cerr<<"Can't interpret creation time"<<endl; return;
  };
  created_b=true;
}

bool SEAttributes::created_compare(const char* c) {
  if(!c) return false;
  string s(c);
  struct tm t;
  if(stringtotime(t,s)) return false;
  return (memcmp(&t,&created_i,sizeof(struct tm)) == 0); 
}

int SEAttributes::write(const char* fname) {
cerr<<"SEAttributes::write: "<<fname<<endl;
  if(!valid) return -1; 
cerr<<"SEAttributes::write: valid"<<endl;
  ofstream o(fname,ios::trunc);
  if(!o) return -1;
cerr<<"SEAttributes::write: opened"<<endl;
  string tmp = id_i; make_escaped_string(tmp);
  o<<"id "<<tmp<<endl;
  o<<"size "<<size_i<<endl;
  tmp = creator_i; make_escaped_string(tmp);
  o<<"creator "<<tmp<<endl;
  if(checksum_b) {
    o<<"checksum "<<checksum_i<<endl;
  };
  if(created_b) {
    o<<"created ";
    int w = o.width(4); int c = o.fill('0');
    o<<(created_i.tm_year+1900);
    o.width(2);
    o<<created_i.tm_mon;
    o<<created_i.tm_mday;
    o<<created_i.tm_hour;
    o<<created_i.tm_min;
    o<<created_i.tm_sec;
    o.width(w); o.fill(c);
    o<<endl;
  };
cerr<<"SEAttributes::write: written"<<endl;
  if(!o) return -1;
cerr<<"SEAttributes::write: exiting"<<endl;
  return 0;
}

bool SEAttributes::enough(void) {
  return (valid && created_b && size_b && id_i.length() && creator_i.length());
}

bool SEAttributes::complete(void) {
cerr<<"SEAttributes::complete: valid: "<<valid<<endl;
cerr<<"SEAttributes::complete: created: "<<created_b<<endl;
cerr<<"SEAttributes::complete: size: "<<size_b<<endl;
cerr<<"SEAttributes::complete: id: "<<id_i<<endl;
cerr<<"SEAttributes::complete: creator: "<<creator_i<<endl;

  return (valid && created_b && size_b && id_i.length() && creator_i.length() && checksum_b);
}

//void SEAttributes::md5_string(string &s) {
//}

