#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <algorithm>
#include <ctype.h>

#include "MdsQuery.h"
#include "LdapQuery.h"
#include "Giis.h"
#include "DateTime.h"
#include "Environment.h"
#include "Utils.h"


// need to wrap for gcc 3.2
inline int to_lower (int x) { return tolower(x); }


struct callback_type {
  vector <Giis> * giislist;
  vector <Cluster> * clusterlist;
};


Job::Job (const string & id_) : id           (id_),
				reqcput      (UNDEFINED),
				queuerank    (UNDEFINED),
				usedcputime  (UNDEFINED),
				usedwalltime (UNDEFINED),
				usedmem      (UNDEFINED),
				cpucount     (UNDEFINED) {}


const string & Job::GetId () const { return id; }

const string & Job::GetStatus () const { return status; }

const string & Job::GetName () const { return jobname; }

const string & Job::GetStdout () const { return stdout; }

const string & Job::GetStderr () const { return stderr; }

const string & Job::GetGmlog () const { return gmlog; }


void Job::SetAttr (const string & attr, const string & value) {

  if (attr == "nordugrid-job-globalid")
    id = value;
  else if (attr == "nordugrid-job-globalowner")
    owner = value;
  else if (attr == "nordugrid-job-execcluster")
    execcluster = value;
  else if (attr == "nordugrid-job-execqueue")
    execqueue = value;
  else if (attr == "nordugrid-job-stdout")
    stdout = value;
  else if (attr == "nordugrid-job-stderr")
    stderr = value;
  else if (attr == "nordugrid-job-stdin")
    stdin = value;
  else if (attr == "nordugrid-job-reqcput")
    reqcput = atoi (value.c_str());
  else if (attr == "nordugrid-job-status")
    status = value;
  else if (attr == "nordugrid-job-queuerank")
    queuerank = atoi (value.c_str());
  else if (attr == "nordugrid-job-lrmscomment")
    lrmscomment = value;
  else if (attr == "nordugrid-job-submissionui")
    submissionui = value;
  else if (attr == "nordugrid-job-submissiontime")
    submissiontime = value;
  else if (attr == "nordugrid-job-usedcputime")
    usedcputime = atoi (value.c_str());
  else if (attr == "nordugrid-job-usedwalltime")
    usedwalltime = atoi (value.c_str());
  else if (attr == "nordugrid-job-sessiondirerasetime")
    erasetime = value;
  else if (attr == "nordugrid-job-usedmem")
    usedmem = atoi (value.c_str());
  else if (attr == "nordugrid-job-errors")
    errors = value;
  else if (attr == "nordugrid-job-jobname")
    jobname = value;
  else if (attr == "nordugrid-job-runtimeenvironment")
    runtimeenv.push_back (value);
  else if (attr == "nordugrid-job-cpucount")
    cpucount = atoi (value.c_str());
  else if (attr == "nordugrid-job-executionnodes")
    execnodes.push_back (value);
  else if (attr == "nordugrid-job-gmlog")
    gmlog = value;
}


void Job::Print (ostream & s) const {

  s << "Job " << id << endl;
  if (!jobname.empty())
    s << "  Jobname: " << jobname << endl;
  s << "  Owner: " << owner << endl;
  int pos = status.find(" at: ");
  if (pos != string::npos) {
    s << "  Status: " << status.substr (0, pos) << ' '
      << UserTime (status.substr (pos + 5, 15)) << endl;
  }
  else {
    s << "  Status: " << status << endl;
  }
  if (!errors.empty())
    s << "  Error: " << errors << endl;
  if (!lrmscomment.empty())
    s << "  Comment: " << lrmscomment << endl;
  if (!execcluster.empty())
    s << "  Cluster: " << execcluster << endl;
  if (!execqueue.empty())
    s << "  Queue: " << execqueue << endl;
  if (cpucount != UNDEFINED)
    s << "  Requested number of CPU:s: " << cpucount << endl;
  if (!execnodes.empty()) {
    s << "  Execution nodes:" << endl;
    for (vector <string>::const_iterator it = execnodes.begin();
	 it != execnodes.end(); it++)
      s << "    " << *it << endl;
  }
  if (queuerank != UNDEFINED)
    s << "  Rank: " << queuerank << endl;
  if (!stdin.empty())
    s << "  stdin: " << stdin << endl;
  if (!stdout.empty())
    s << "  stdout: " << stdout << endl;
  if (!stderr.empty())
    s << "  stderr: " << stderr << endl;
  if (!gmlog.empty())
    s << "  Grid manager log directory: " << gmlog << endl;
  if (!submissionui.empty())
    s << "  Submitted from: " << submissionui << endl;
  if (!submissiontime.empty())
    s << "  Submitted: " << UserTime (submissiontime) << endl;
  if (!runtimeenv.empty()) {
    s << "  Requested runtime environments:" << endl;
    for (vector <Environment>::const_iterator it = runtimeenv.begin();
	 it != runtimeenv.end(); it++)
      s << "    " << *it << endl;
  }
  if (reqcput != UNDEFINED)
    s << "  Required CPU time: " << Period (reqcput) << endl;
  if (usedcputime != UNDEFINED)
    s << "  Used CPU time: " << Period (usedcputime) << endl;
  if (usedwalltime != UNDEFINED)
    s << "  Used wall time: " << Period (usedwalltime) << endl;
  if (usedmem != UNDEFINED)
    s << "  Used memory: " << usedmem << " kB" << endl;
  if (!erasetime.empty())
    s << "  Results " << (status == "DELETED" ? "should have been" : "must be")
      << " retrieved before: " << UserTime (erasetime) << endl;
}

void Job::PrintShort (ostream & s) const {

  s << "Job " << id << endl;
  if (!jobname.empty())
    s << "  Jobname: " << jobname << endl;
  int pos = status.find(" at: ");
  if (pos != string::npos) {
    s << "  Status: " << status.substr (0, pos) << ' '
      << UserTime (status.substr (pos + 5, 15)) << endl;
  }
  else {
    s << "  Status: " << status << endl;
  }
  if (!errors.empty())
    s << "  Error: " << errors << endl;
}


bool Job::IsStatus (const vector <string> & stat) {

  if (stat.empty())
    return true;
  for (vector <string>::const_iterator vsci = stat.begin();
       vsci != stat.end(); vsci++) {
    if (!errors.empty()) {
      if (strcasecmp (vsci->c_str(), "FAILED") == 0)
	return true;
    }
    else {
      int pos = status.find (" at: ");
      if (pos == string::npos) {
        if (strcasecmp (vsci->c_str(), status.c_str()) == 0)
	  return true;
      }
      else {
	if (strcasecmp (vsci->c_str(), status.substr (0, pos).c_str()) == 0)
	  return true;
      }
    }
  }
  return false;
}


Queue::Queue (const string & name_) : name              (name_),
				      running           (UNDEFINED),
				      queued            (UNDEFINED),
				      maxrunning        (UNDEFINED),
				      maxqueuable       (UNDEFINED),
				      maxuserrun        (UNDEFINED),
				      maxcputime        (UNDEFINED),
				      mincputime        (UNDEFINED),
				      defaultcputime    (UNDEFINED),
				      totalcpus         (UNDEFINED),
				      nodemem           (UNDEFINED),
				      gridrunning       (UNDEFINED),
				      gridqueued        (UNDEFINED),
				      userdiskspace     (UNDEFINED),
				      userqueuelength   (UNDEFINED) {}


const string & Queue::GetName () const { return name; }

const string & Queue::GetStatus () const { return status; }

int Queue::GetRunning () const { return running; }

int Queue::GetQueued () const { return queued; }

int Queue::GetMaxRunning () const { return maxrunning; }

int Queue::GetMaxQueuable () const { return maxqueuable; }

int Queue::GetMaxCpuTime () const { return maxcputime; }

int Queue::GetMinCpuTime () const { return mincputime; }

int Queue::GetDefaultCpuTime () const { return defaultcputime; }

int Queue::GetTotalCpus () const { return totalcpus; }

int Queue::GetNodeMemory () const { return nodemem; }

const string & Queue::GetArchitecture () const { return arch; }


int Queue::GetFrequency () const {

  if (nodecpu.empty()) return UNDEFINED;
  int pos = nodecpu.find (" MHz");
  if (pos == string::npos) {
    cerr << "Warning: Bad format for nodecpu queue MDS attribute." << endl;
    return UNDEFINED;
  }
  int numpos = nodecpu.rfind (' ', pos - 1) + 1;
  return atoi (nodecpu.substr (numpos, pos - numpos).c_str());
}


int Queue::GetUserFreeCpus (int minutes) const {

  if (userfreecpus.empty()) return UNDEFINED;
  if (minutes == UNDEFINED) minutes == INT_MAX;
  map <int, int>::const_iterator it = userfreecpus.lower_bound (minutes);
  if (it == userfreecpus.end()) return 0;
  return it->second;
}


long long int Queue::GetUserDiskSpace () const { return userdiskspace; }

void Queue::ClaimUserDiskSpace (long long int size) { userdiskspace -= size; }


const Environment * Queue::FindEnvironment (const string & attr,
					    const EnvironmentTest &
					    envtest) const {

  string lcattr (attr.length(), '\0');
  transform (attr.begin(), attr.end(), lcattr.begin(), to_lower);
  const Environment * env = NULL;
  if (lcattr == "middleware") {
    for (vector <Environment>::const_iterator vei = middleware.begin();
	 vei != middleware.end(); vei++)
      if (envtest.Test (*vei))
	if (!env || (env->GetVersion() < vei->GetVersion())) env = &*vei;
  }
  else if (lcattr == "runtimeenvironment") {
    for (vector <Environment>::const_iterator vei = runtimeenv.begin();
	 vei != runtimeenv.end(); vei++)
      if (envtest.Test (*vei))
	if (!env || (env->GetVersion() < vei->GetVersion())) env = &*vei;
  }
  else if (lcattr == "opsys") {
    for (vector <Environment>::const_iterator vei = opsys.begin();
	 vei != opsys.end(); vei++)
      if (envtest.Test (*vei))
	if (!env || (env->GetVersion() < vei->GetVersion())) env = &*vei;
  }
  else
    cerr << "Unexpected attribute in FindEnvironment - should never happen"
	 << endl;
  return env;
}


void Queue::SetAttr (const string & attr, const string & value) {

  if (attr == "nordugrid-queue-name")
    name = value;
  else if (attr == "nordugrid-queue-status")
    status = value;
  else if (attr == "nordugrid-queue-running")
    running = atoi (value.c_str());
  else if (attr == "nordugrid-queue-queued")
    queued = atoi (value.c_str());
  else if (attr == "nordugrid-queue-maxrunning")
    maxrunning = atoi (value.c_str());
  else if (attr == "nordugrid-queue-maxqueuable")
    maxqueuable = atoi (value.c_str());
  else if (attr == "nordugrid-queue-maxuserrun")
    maxuserrun = atoi (value.c_str());
  else if (attr == "nordugrid-queue-maxcputime")
    maxcputime = atoi (value.c_str());
  else if (attr == "nordugrid-queue-mincputime")
    mincputime = atoi (value.c_str());
  else if (attr == "nordugrid-queue-defaultcputime")
    defaultcputime = atoi (value.c_str());
  else if (attr == "nordugrid-queue-schedulingpolicy")
    schedpolicy = value;
  else if (attr == "nordugrid-queue-totalcpus")
    totalcpus = atoi (value.c_str());
  else if (attr == "nordugrid-queue-nodecpu")
    nodecpu = value;
  else if (attr == "nordugrid-queue-nodememory")
    nodemem = atoi (value.c_str());
  else if (attr == "nordugrid-queue-architecture")
    arch = value;
  else if (attr == "nordugrid-queue-opsys")
    opsys.push_back (value);
  else if (attr == "nordugrid-queue-gridrunning")
    gridrunning = atoi (value.c_str());
  else if (attr == "nordugrid-queue-gridqueued")
    gridqueued = atoi (value.c_str());
  else if (attr == "nordugrid-queue-runtimeenvironment")
    runtimeenv.push_back (value);
  else if (attr == "nordugrid-queue-middleware")
    middleware.push_back (value);
  else if (attr == "nordugrid-authuser-freecpus")
    userfreecpus = ParseStringToMap (value, true);
  else if (attr == "nordugrid-authuser-diskspace")
    userdiskspace = atoll (value.c_str()) * 1024 * 1024;
  else if (attr == "nordugrid-authuser-queuelength")
    userqueuelength = atoi (value.c_str());
}


void Queue::Print (ostream & s) const {

  s << "Queue " << name << endl;
  s << "  Status: " << status << endl;
  if (!arch.empty())
    s << "  Architecture: " << arch << endl;
  if (!opsys.empty())
    s << "  Operating system:" << endl;
    for (vector <Environment>::const_iterator it = opsys.begin();
	 it != opsys.end(); it++)
      s << "    " << *it << endl;
  if (!nodecpu.empty())
    s << "  CPU: " << nodecpu << endl;
  if (nodemem != UNDEFINED)
    s << "  Memory on each node: " << nodemem << " MB" << endl;
  if (!middleware.empty()) {
    s << "  Installed middleware:" << endl;
    for (vector <Environment>::const_iterator it = middleware.begin();
	 it != middleware.end(); it++)
      s << "    " << *it << endl;
  }
  if (!runtimeenv.empty()) {
    s << "  Installed runtime environments:" << endl;
    for (vector <Environment>::const_iterator it = runtimeenv.begin();
	 it != runtimeenv.end(); it++)
      s << "    " << *it << endl;
  }
  if (running != UNDEFINED)
    s << "  Number of running jobs: " << running << endl;
  if (queued != UNDEFINED)
    s << "  Number of queued jobs: " << queued << endl;
  if (maxrunning != UNDEFINED)
    s << "  Max number of running jobs: " << maxrunning << endl;
  if (maxqueuable != UNDEFINED)
    s << "  Max number of queued jobs: " << maxqueuable << endl;
  if (maxuserrun != UNDEFINED)
    s << "  Max number of running jobs per local user: " << maxuserrun << endl;
  if (maxcputime != UNDEFINED)
    s << "  Max CPU time: " << Period (maxcputime) << endl;
  if (mincputime != UNDEFINED)
    s << "  Min CPU time: " << Period (mincputime) << endl;
  if (defaultcputime != UNDEFINED)
    s << "  Default CPU time: " << Period (defaultcputime) << endl;
  if (!schedpolicy.empty())
    s << "  Scheduling policy: " << schedpolicy << endl;
  if (totalcpus != UNDEFINED)
    s << "  Number of CPUs: " << totalcpus << endl;
  if (!userfreecpus.empty()) {
    s << "  User number of free CPUs:" << endl;
    for (map <int, int>::const_iterator it = userfreecpus.begin();
	 it != userfreecpus.end(); it++)
      if (it->second)
	s << "    " << it->second
	  << (it->second == 1 ? " processor for " : " processors for ")
	  << (it->first == INT_MAX ? "infinite time" : Period (it->first))
	  << endl;
      else
	s << "    None" << endl;
  }
  if (userdiskspace != UNDEFINED)
    s << "  User disk space: " << userdiskspace << " bytes" << endl;
  if (userqueuelength != UNDEFINED)
    s << "  User queue length: " << userqueuelength << endl;
}


void Queue::PrintShort (ostream & s) const {

  s << "Queue " << name << endl;
  s << "  Status: " << status << endl;
}


void Queue::Accept (int minutes, int count) {

  if (running < maxrunning)
    running++;
  else
    queued++;

  if (minutes == UNDEFINED) minutes = INT_MAX;
  map <int, int>::iterator last = userfreecpus.lower_bound (minutes);
  if (last != userfreecpus.end() && last->second >= count) {
    for (map <int, int>::iterator it = userfreecpus.begin();
	 it != userfreecpus.end(); it++) {
      if (it->first <= last->first)
	userfreecpus [it->first] -= count;
      else if (it->second >= last->second) {
	userfreecpus [it->first] = last->second;
	int oldkey = last->first;
	last++;
	userfreecpus.erase (oldkey);
      }
    }
    if (last->second == 0) userfreecpus.erase (last->first);
    if (userfreecpus.empty())
      userfreecpus [(maxcputime != UNDEFINED ? maxcputime : INT_MAX)] = 0;
  }
}


Cluster::Cluster (const string & name_) : name        (name_),
					  homogeneity (false),
					  nodemem     (UNDEFINED),
					  totalcpus   (UNDEFINED),
					  totaljobs   (UNDEFINED),
					  usedcpus    (UNDEFINED),
					  queuedjobs  (UNDEFINED),
					  sessiondirfree  (UNDEFINED),
					  sessiondirtotal (UNDEFINED),
					  cachefree   (UNDEFINED),
					  cachetotal  (UNDEFINED) {}


const string & Cluster::GetName () const { return name; }


const string Cluster::GetLongName () const {

  if (!alias.empty())
    return alias + " (" + name + ")";
  else
    return name;
}


const string & Cluster::GetContact () const { return contact; }

const string & Cluster::GetArchitecture () const { return arch; }


int Cluster::GetFrequency () const {
  
  if (nodecpu.empty()) return UNDEFINED;
  int pos = nodecpu.find (" MHz");
  if (pos == string::npos) {
    cerr << "Warning: Bad format for nodecpu cluster MDS attribute." << endl;
    return UNDEFINED;
  }
  int numpos = nodecpu.rfind (' ', pos - 1) + 1;
  return atoi (nodecpu.substr (numpos, pos - numpos).c_str());
}


int Cluster::GetNodeMemory () const { return nodemem; }

int Cluster::GetTotalCpus () const { return totalcpus; }

int Cluster::GetQueued () const { return queuedjobs; }

long long int Cluster::GetCacheFree () const { return cachefree; }

void Cluster::ClaimCache (long long int size) { cachefree -= size; }

long long int Cluster::GetSessionDirFree () const { return sessiondirfree; }

void Cluster::ClaimSessionDir (long long int size) { sessiondirfree -= size; }


const Environment * Cluster::FindEnvironment (const string & attr,
					      const EnvironmentTest &
					      envtest) const {

  string lcattr (attr.length(), '\0');
  transform (attr.begin(), attr.end(), lcattr.begin(), to_lower);
  const Environment * env = NULL;
  if (lcattr == "middleware") {
    for (vector <Environment>::const_iterator vei = middleware.begin();
	 vei != middleware.end(); vei++)
      if (envtest.Test (*vei))
	if (!env || (env->GetVersion() < vei->GetVersion())) env = &*vei;
  }
  else if (lcattr == "runtimeenvironment") {
    for (vector <Environment>::const_iterator vei = runtimeenv.begin();
	 vei != runtimeenv.end(); vei++)
      if (envtest.Test (*vei))
	if (!env || (env->GetVersion() < vei->GetVersion())) env = &*vei;
  }
  else if (lcattr == "opsys") {
    for (vector <Environment>::const_iterator vei = opsys.begin();
	 vei != opsys.end(); vei++)
      if (envtest.Test (*vei))
	if (!env || (env->GetVersion() < vei->GetVersion())) env = &*vei;
  }
  else
    cerr << "Unexpected attribute in FindEnvironment - should never happen"
	 << endl;
  return env;
}


bool Cluster::MatchLocalSe (const string & location) const {

  string loc = location;
  RemoveDefaultPort (loc);
  for (vector <string>::const_iterator vsi = localse.begin();
       vsi != localse.end(); vsi++) {
    string lse = *vsi;
    RemoveDefaultPort (lse);
    if (lse [lse.length() - 1] == '/') lse.erase (lse.length() - 1);
    if (lse == loc) return true;
    if (lse == loc.substr (0, lse.length()) && loc [lse.length()] == '/')
      return true;
  }
  return false;
}


bool Cluster::HaveNodeAccess (const string & attr) const {

  for (vector <string>::const_iterator vsi = nodeaccess.begin();
       vsi != nodeaccess.end(); vsi++)
    if (*vsi == attr) return true;
  return false;
}


void Cluster::SetAttr (const string & attr, const string & value) {

  if (attr == "nordugrid-cluster-name")
    name = value;
  else if (attr == "nordugrid-cluster-aliasname")
    alias = value;
  else if (attr == "nordugrid-cluster-contactstring")
    contact = value;
  else if (attr == "nordugrid-cluster-support")
    support.push_back (value);
  else if (attr == "nordugrid-cluster-lrms-type")
    lrmstype = value;
  else if (attr == "nordugrid-cluster-lrms-version")
    lrmsversion = value;
  else if (attr == "nordugrid-cluster-lrms-config")
    lrmsconfig = value;
  else if (attr == "nordugrid-cluster-architecture")
    arch = value;
  else if (attr == "nordugrid-cluster-opsys")
    opsys.push_back (value);
  else if (attr == "nordugrid-cluster-homogeneity")
    homogeneity = (value == "True");
  else if (attr == "nordugrid-cluster-nodecpu")
    nodecpu = value;
  else if (attr == "nordugrid-cluster-nodememory")
    nodemem = atoi (value.c_str());
  else if (attr == "nordugrid-cluster-totalcpus")
    totalcpus = atoi (value.c_str());
  else if (attr == "nordugrid-cluster-cpudistribution")
    cpudist = ParseStringToMap (value);
  else if (attr == "nordugrid-cluster-sessiondir-free")
    sessiondirfree = atoll (value.c_str()) * 1024 * 1024;
  else if (attr == "nordugrid-cluster-sessiondir-total")
    sessiondirtotal = atoll (value.c_str()) * 1024 * 1024;
  else if (attr == "nordugrid-cluster-cache-free")
    cachefree = atoll (value.c_str()) * 1024 * 1024;
  else if (attr == "nordugrid-cluster-cache-total")
    cachetotal = atoll (value.c_str()) * 1024 * 1024;
  else if (attr == "nordugrid-cluster-runtimeenvironment")
    runtimeenv.push_back (value);
  else if (attr == "nordugrid-cluster-localse")
    localse.push_back (value);
  else if (attr == "nordugrid-cluster-middleware")
    middleware.push_back (value);
  else if (attr == "nordugrid-cluster-totaljobs")
    totaljobs = atoi (value.c_str());
  else if (attr == "nordugrid-cluster-usedcpus")
    usedcpus = atoi (value.c_str());
  else if (attr == "nordugrid-cluster-queuedjobs")
    queuedjobs = atoi (value.c_str());
  else if (attr == "nordugrid-cluster-location")
    location = value;
  else if (attr == "nordugrid-cluster-owner")
    owner.push_back (value);
  else if (attr == "nordugrid-cluster-issuerca")
    issuer = value;
  else if (attr == "nordugrid-cluster-nodeaccess")
    nodeaccess.push_back (value);
  // kept for backward compatibility
  else if (attr == "nordugrid-cluster-opsysdistribution")
    opsys.push_back (value);
}


void Cluster::Print (ostream & s) const {

  s << "Cluster " << name << endl;
  if (!alias.empty())
    s << "  Alias: " << alias << endl;
  if (!location.empty())
    s << "  Location: " << location << endl;
  if (!issuer.empty()) {
    s << "  Host certificate issuer:" << endl;
    s << "    " << issuer << endl;
  }
  if (!owner.empty()) {
    s << "  Owners:" << endl;
    for (vector <string>::const_iterator it = owner.begin();
	 it != owner.end(); it++)
      s << "    " << *it << endl;
  }
  if (!arch.empty())
    s << "  Architecture: " << arch << endl;
  if (!opsys.empty())
    s << "  Operating system:" << endl;
    for (vector <Environment>::const_iterator it = opsys.begin();
	 it != opsys.end(); it++)
      s << "    " << *it << endl;
  if (!nodecpu.empty())
    s << "  CPU" << (homogeneity ? ": " : " (worst): ") << nodecpu << endl;
  if (nodemem != UNDEFINED)
    s << "  Memory on each node" << (homogeneity ? ": " : " (smallest): ")
      << nodemem << " MB" << endl;
  if (sessiondirtotal != UNDEFINED)
    s << "  Size of scratch dir: " << sessiondirtotal << " bytes" << endl;
  if (sessiondirfree != UNDEFINED)
    s << "  Free space in scratch dir: " << sessiondirfree << " bytes" << endl;
  if (cachetotal != UNDEFINED)
    s << "  Size of cache dir: " << cachetotal << " bytes" << endl;
  if (cachefree != UNDEFINED)
    s << "  Free space in cache dir: " << cachefree << " bytes" << endl;
  if (totalcpus != UNDEFINED)
    s << "  Number of CPUs: " << totalcpus << endl;
  if (usedcpus != UNDEFINED)
    s << "  Number of used CPUs: " << usedcpus << endl;
  if (totaljobs != UNDEFINED)
    s << "  Number of jobs: " << totaljobs << endl;
  if (queuedjobs != UNDEFINED)
    s << "  Number of queued jobs: " << queuedjobs << endl;
  if (!cpudist.empty()) {
    s << "  CPU distribution:" << endl;
    for (map <int, int>::const_iterator it = cpudist.begin();
	 it != cpudist.end(); it++) {
      s << "    ";
      if      (it->first == 1) s << "single";
      else if (it->first == 2) s << "dual";
      else                     s << it->first;
      s << " CPU machines: " << it->second << endl;
    }
  }
  if (!support.empty()) {
    s << "  Cluster support:" << endl;
    for (vector <string>::const_iterator it = support.begin();
	 it != support.end(); it++)
      s << "    " << *it << endl;
  }
  if (!localse.empty()) {
    s << "  Local Storage Elements:" << endl;
    for (vector <string>::const_iterator it = localse.begin();
	 it != localse.end(); it++)
      s << "    " << *it << endl;
  }
  if (!middleware.empty()) {
    s << "  Installed middleware:" << endl;
    for (vector <Environment>::const_iterator it = middleware.begin();
	 it != middleware.end(); it++)
      s << "    " << *it << endl;
  }
  if (!runtimeenv.empty()) {
    s << "  Installed runtime environments:" << endl;
    for (vector <Environment>::const_iterator it = runtimeenv.begin();
	 it != runtimeenv.end(); it++)
      s << "    " << *it << endl;
  }
  if (!nodeaccess.empty()) {
    s << "  Network access on cluster nodes:" << endl;
    for (vector <string>::const_iterator it = nodeaccess.begin();
	 it != nodeaccess.end(); it++)
      s << "    " << *it << endl;
  }
}


void Cluster::PrintShort (ostream & s) const {

  s << "Cluster " << name << endl;
  if (!alias.empty())
  s << "  Alias: " << alias << endl;
}


void FindInfoCallback (const string & attr,
		       const string & value, void * ref) {

  Cluster * c = (Cluster *) ref;

  static int qix;
  static int jix;

  string lcattr (attr.length(), '\0');
  transform (attr.begin(), attr.end(), lcattr.begin(), to_lower);
  if (lcattr == "dn") {
    string dn (value.length(), '\0');
    transform (value.begin(), value.end(), dn.begin(), to_lower);
    int pos;
    qix = UNDEFINED;
    pos = dn.find ("nordugrid-queue-name");
    if (pos != string::npos) {
      pos = dn.find ('=', pos);
      string qname = value.substr (pos + 1, dn.find (',', pos + 1) - pos - 1);
      while (qname [qname.length() - 1] == ' ')
	qname.erase (qname.length() - 1);
      for (int i = 0; i < c->queues.size(); i++)
	if (qname == c->queues[i].GetName()) {qix = i; break;}
      if (qix == UNDEFINED) {
	qix = c->queues.size();
	c->queues.push_back (qname);
      }
    }
    jix = UNDEFINED;
    pos = dn.find ("nordugrid-job-globalid");
    if (pos != string::npos) {
      pos = dn.find ('=', pos);
      string jobid = value.substr (pos + 1, dn.find (',', pos + 1) - pos - 1);
      while (jobid [jobid.length() - 1] == ' ')
	jobid.erase (jobid.length() - 1);
      for (int i = 0; i < c->queues[qix].jobs.size(); i++)
	if (jobid == c->queues[qix].jobs[i].GetId()) {jix = i; break;}
      if (jix == UNDEFINED) {
	jix = c->queues[qix].jobs.size();
	c->queues[qix].jobs.push_back (jobid);
      }
    }
  }
  else {
    if (qix == UNDEFINED)
      c->SetAttr (lcattr, value);
    else if (jix == UNDEFINED)
      c->queues[qix].SetAttr (lcattr, value);
    else
      c->queues[qix].jobs[jix].SetAttr (lcattr, value);
  }
}


int Cluster::Connect (const string & usersn, bool anonymous,
		      int timeout, int debug) {

  return ldap.Connect (name, 2135, usersn, anonymous, timeout, debug);
}


int Cluster::Query (const string & usersn, int timeout, int debug) {

  const string filter = "(|(objectclass=nordugrid-cluster)(objectclass=nordugrid-queue)(nordugrid-authuser-sn=" + usersn + ")(nordugrid-job-globalowner=" + usersn + "))";
  const vector <string> attributes;

  return ldap.Query ("Mds-Vo-name=local,o=grid", filter, attributes,
		     LdapQuery::subtree, timeout, debug);
}


int Cluster::Result (int timeout, int debug) {

  return ldap.Result (FindInfoCallback, this, timeout, debug);
}


int Cluster::Find (const string & usersn, bool anonymous,
		   int timeout, int debug) {

  if (Connect (usersn, anonymous, timeout, debug)) return 1;
  if (Query (usersn, timeout, debug)) return 1;
  return Result (timeout, debug);
}


void FindClusterInfo (vector <Cluster> & clusterlist,
		      const string & usersn,
		      bool anonymous,
		      int timeout,
		      int debug) {

  for (vector <Cluster>::iterator cli = clusterlist.begin();
       cli != clusterlist.end(); cli++)
    cli->Connect (usersn, anonymous, timeout, debug);
  for (vector <Cluster>::iterator cli = clusterlist.begin();
       cli != clusterlist.end(); cli++)
    cli->Query (usersn, timeout, debug);
  for (vector <Cluster>::iterator cli = clusterlist.begin();
       cli != clusterlist.end(); cli++)
    cli->Result (timeout, debug);
}


void FindClustersCallback (const string & attr,
			   const string & value, void * ref) {

  struct callback_type * cb = (struct callback_type *) ref;

  static string host;
  static int port;
  static string suffix;
  static bool newgiis;
  static bool newcluster;

  string lcattr (attr.length(), '\0');
  transform (attr.begin(), attr.end(), lcattr.begin(), to_lower);
  if (lcattr == "mds-service-hn") {
    host = value;
    newcluster = false;
    newgiis = false;
  }
  else if (lcattr == "mds-service-port") {
    port = atoi (value.c_str());
  }
  else if (lcattr == "mds-service-ldap-suffix") {
    string lcvalue (value.length(), '\0');
    transform (value.begin(), value.end(), lcvalue.begin(), to_lower);
    if (lcvalue.substr (0, 17) == "mds-vo-name=local" ||
	lcvalue.substr (0, 22) == "nordugrid-cluster-name") {
      newcluster = true;
    }
    else if (lcvalue.substr (0, 11) == "mds-vo-name") {
      newgiis = true;
      suffix = value;
    }
  }
  else if (lcattr == "mds-reg-status") {
    if (value == "VALID")
      if (newcluster) {
	bool old = false;
	for (vector <Cluster>::iterator cli = cb->clusterlist->begin();
	     !old && cli != cb->clusterlist->end(); cli++)
	  if (cli->GetName() == host) old = true;
	if (!old) cb->clusterlist->push_back (host);
      }
      else if (newgiis) {
	Giis giis (host, port, suffix);
	bool old = false;
	for (vector <Giis>::iterator vgi = cb->giislist->begin();
	     !old && vgi != cb->giislist->end(); vgi++) {
	  if (*vgi == giis) old = true;
	}
	if (!old) cb->giislist->push_back (giis);
      }
  }
}


vector <Cluster> FindClusters (vector <Giis> giislist,
			       const string & usersn,
			       bool anonymous,
			       int timeout,
			       int debug) {

  callback_type cb;
  vector <Cluster> cl;
  vector <Giis> gl = giislist;
  cb.giislist = &gl;
  cb.clusterlist = &cl;

  int nqueried = 0;

  while (nqueried < giislist.size()) {
    int last = giislist.size();
    for (int i = nqueried; i < last; i++)
      giislist [i].Connect (usersn, anonymous, timeout, debug);
    for (int i = nqueried; i < last; i++)
      giislist [i].Query (usersn, timeout, debug);
    for (int i = nqueried; i < last; i++)
      giislist [i].Result (FindClustersCallback, &cb, timeout, debug);
    giislist = gl;
    nqueried = last;
  }

  return cl;
}
